{-# LANGUAGE GeneralizedNewtypeDeriving, FlexibleInstances #-}
{-# OPTIONS_GHC -Wall -fno-warn-incomplete-patterns -fno-warn-name-shadowing #-}
module Lam.TextRenderer(
  render
) where

import Control.Applicative
import Control.Monad.Instances
import Data.Char
import Data.Maybe
import Lam.Lam
import Lam.Adapters
import Lam.Binding
import Data.Cursor.CLASE.Language
import Data.Cursor.CLASE.Traversal
import qualified Data.Map as Map

render :: Cursor Lam x a -> String
render = (\(s,()) -> s) . ($ Map.empty) . fromLBI . fromTR . completeTraversal

newtype TextRenderer x = TR { fromTR :: (LamBinderImpl (String, x)) }

instance Functor TextRenderer where
  fmap f (TR x) = TR (fmap (fmap f) $ x)

instance Applicative TextRenderer where
  (TR f) <*> (TR l) = TR (fmap (uncurry (<*>)) (liftA2 (,) f l))
  pure v = TR (pure (pure v))

instance Monad TextRenderer where
  return = pure
  (TR f) >>= fn = TR (f >>= fromTR . fn . snd)

instance LamBinder TextRenderer where
  addBinding s (TR x) = TR (addBinding s x)
  lookupBinding i = TR (fmap pure $ lookupBinding i)
  getBindingMap = TR (fmap pure $ getBindingMap)

string :: String -> TextRenderer ()
string s = TR (pure (s,()))

space :: TextRenderer ()
space = string " "

instance LamTraversalAdapterLam (TextRenderer ()) where
  visitLam _ hole = hole

instance LamTraversalAdapterExp (TextRenderer ()) where
  visitAbs (Abs name _ _) ty exp = string ("λ " ++ name ++ "∷") *> ty *> space *> exp
  visitApp _ l r = l *> string "∘" *> r
  visitVar (Var i) = (string . fromMaybe "Variable free!" =<< lookupBinding i) *> (string . subscript $ i)

instance LamTraversalAdapterType (TextRenderer ()) where
  visitUnit _ = string "τ"
  visitArr _ lhs rhs = lhs *> string " → " *> rhs

instance LamTraversalAdapterCursor (TextRenderer ()) where
  visitCursor _ child = string "⟦" *> child *> string "⟧"

subscript :: Integer -> String
subscript = map (chr . (+) (8320 - (ord '0')) . ord) . show 
